/* Supplimentary functions that get compiled and linked to cpp on an
   Amiga host.
   Copyright (C) 1996 Free Software Foundation, Inc.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#if HAVE_STDLIB_H
# include <stdlib.h>
#else
char *getenv ();
#endif

#if STDC_HEADERS
# include <string.h>
#else
char *strrchr();
#endif

#include <proto/exec.h>
#include <proto/dos.h>

static int amiga_priority = -1;	 /* Flag indicating process priority */
struct Task *amiga_task;
BPTR __amiga_filehandle (int fd); /* Semi-private function of IXEmul. */

/* Improved version of open() that also checks if casing of FILENAME is
   correct. */

#ifndef O_CASE	/* Use this until ixemul.library implements O_CASE */

int
amiga_open_case_sensitive (filename, flags, mode)
     char *filename;
     int flags;
     mode_t mode;
{
  int fd;
  BPTR fh;
  char fibbuf[sizeof (struct FileInfoBlock) + 3], *basename;
  struct FileInfoBlock *fib;

  fd = open (filename, flags, mode);
  if (fd != -1)
    {
      /* Ensure that casing of the filename is correct. */

      fh = __amiga_filehandle (fd);
      /* Align fib to longword. */
      fib = (struct FileInfoBlock*) ((ULONG) (fibbuf + 3) & 0xFFFFFFFC);
      if (!(basename=strrchr(filename, '/')))
	basename=filename;
      else
	basename++; /* Skip '/'. */

      /* The last condition below is to prevent incorrect handling of links. */
      if (fh && ExamineFH (fh, fib) && strcmp (fib -> fib_FileName, basename)
          && strcasecmp (fib -> fib_FileName, basename) == 0)
	{
	  close (fd);
	  errno = ENOENT;
	  fd = -1;
	}
    }
  return (fd);
}

#endif	/* #ifndef O_CASE */

/* This function returns whether the LEN characters long filename FNAME 
   is an absolute path specification. */

int
amiga_abs_filename (fname, len)
     char *fname;
     int len;
{
  /* we're using ixemul.library, which treats `/foo' as `foo:', so 
     fname[0] is to be considered absolute as well */

  if (fname[0] == '/')
    {
      return (1);
    }

  /* else do an index() on fname, but one which is limited to len characters */

  while (*fname && *fname != ':' && len) 
    {
      fname++;
      len--;
    }

  return (*fname == ':');
}

void
amiga_get_default_priority ()
{
  char *envstr;
  int i;

  if (envstr = getenv ("GCCPRIORITY"))
    {
      if (((i = atoi (envstr)) > -20) && (i < 20))
	{
	  amiga_priority = i;
	}
    }
}

/* Set the priority to whatever is currently in the amiga_priority variable. */

void
amiga_set_priority ()
{
  Forbid ();
  amiga_task = FindTask (NULL);
  Permit ();
  SetTaskPri (amiga_task, amiga_priority);
}

/* Parse the --priority option and stash the value away for future use. */

int
amiga_parse_priority (argc, argv, ip)
     int argc;
     char **argv;
     int *ip;
{
  if (strcmp (argv[*ip], "-priority") != 0)
    {
      return (0);
    }
  else
    {
      if (*ip + 1 == argc)
	{
	  fatal ("Priority missing after -P option");
	}
      (*ip)++;
      if ((amiga_priority = atoi (argv[*ip])) < -25)
	{
	  amiga_priority = -25;
	}
      else if (amiga_priority > 25)
	{
	  amiga_priority = 25;
	}
      return (1);
    }
}

/* This function is called with the name of each symbol found in a
   #if, #ifdef, #ifndef, or #if defined directive.  Presently it just
   warns users not to test for obsolete "amigados" or "__amigados__" */

void
amiga_cpp_symname_hook (char *symname, int symlength)
{
  if (symlength == 8 && strncmp(symname, "amigados", symlength) == 0 ||
      symlength == 12 && strncmp(symname, "__amigados__", symlength) == 0)
    {
      char save = *(symname + symlength);
      *(symname + symlength) = '\000';
      warning ("change test for '%s' to use '__amigaos__' instead", symname);
      warning ("this symbol will not be predefined in a future release");
      *(symname + symlength) = save;
    }  
}
